package me.parsell.wireddust.common;

import java.util.Iterator;
import java.util.Map;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.UnmodifiableIterator;

import org.apache.logging.log4j.core.jmx.Server;
import org.jetbrains.annotations.Nullable;

import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.block.ShapeContext;
import net.minecraft.entity.EquipmentSlot;
import net.minecraft.entity.LivingEntity;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.item.ItemPlacementContext;
import net.minecraft.item.ItemStack;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.state.StateManager;
import net.minecraft.state.property.EnumProperty;
import net.minecraft.util.ActionResult;
import net.minecraft.util.BlockMirror;
import net.minecraft.util.BlockRotation;
import net.minecraft.util.Hand;
import net.minecraft.util.hit.BlockHitResult;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.util.shape.VoxelShape;
import net.minecraft.util.shape.VoxelShapes;
import net.minecraft.world.BlockView;
import net.minecraft.world.World;
import net.minecraft.world.WorldAccess;
import net.minecraft.world.WorldView;

public class WDWireBlock extends Block{
	// These are the states of the wires coming off our block
  public static final EnumProperty<WDWireConnection> WIRE_CONNECTION_NORTH;
  public static final EnumProperty<WDWireConnection> WIRE_CONNECTION_SOUTH;
  public static final EnumProperty<WDWireConnection> WIRE_CONNECTION_EAST;
  public static final EnumProperty<WDWireConnection> WIRE_CONNECTION_WEST;

	// Locally defined DIRECTION <-> WIRE_CONNECTION_DIRECTION
	public static final Map<Direction, EnumProperty<WDWireConnection>> DIRECTION_TO_WIRE_CONNECTION;

	// Determines the hitbox for our block
	public static final VoxelShape DOT_VOXELSHAPE;
	private static final Map<Direction, VoxelShape> DIRECTION_VOXELSHAPE, UP_VOXELSHAPE;
	public final Map<BlockState, VoxelShape> STATE_TO_VOXELSHAPE = Maps.newHashMap(); 


	public WDWireBlock(Settings settings) {
		super(settings);
		// Set these properties to be none as start, so we start with a dot
		this.setDefaultState((BlockState)((BlockState)((BlockState)((BlockState)((BlockState)((BlockState)this.stateManager.getDefaultState()).with(WIRE_CONNECTION_NORTH, WDWireConnection.NONE)).with(WIRE_CONNECTION_EAST, WDWireConnection.NONE)).with(WIRE_CONNECTION_SOUTH, WDWireConnection.NONE)).with(WIRE_CONNECTION_WEST, WDWireConnection.NONE)));
		
		UnmodifiableIterator<BlockState> stateItr = this.getStateManager().getStates().iterator();
		while(stateItr.hasNext()) {
			BlockState state = stateItr.next();
			this.STATE_TO_VOXELSHAPE.put(state, determineVoxelShape(state));
		}
	}

	// Make sure out block has the properties for us to manage
	@Override
	protected void appendProperties(StateManager.Builder<Block, BlockState> builder) {
		builder.add(WIRE_CONNECTION_NORTH, WIRE_CONNECTION_EAST, WIRE_CONNECTION_SOUTH, WIRE_CONNECTION_WEST);
	}

	public VoxelShape determineVoxelShape(BlockState state){
		VoxelShape voxelShape = DOT_VOXELSHAPE;

		Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
		while(dirItr.hasNext()) {
		   Direction direction = (Direction)dirItr.next();
		   WDWireConnection wireConnection = state.get(DIRECTION_TO_WIRE_CONNECTION.get(direction));
			if (wireConnection == WDWireConnection.SIDE) {
				voxelShape = VoxelShapes.union(voxelShape, (VoxelShape)DIRECTION_VOXELSHAPE.get(direction));
		    } else if (wireConnection == WDWireConnection.UP) {
				voxelShape = VoxelShapes.union(voxelShape, (VoxelShape)UP_VOXELSHAPE.get(direction));
		    }
		}

		return voxelShape;
	}

	@Override
	public VoxelShape getOutlineShape(BlockState state, BlockView world, BlockPos pos, ShapeContext context) {
		return STATE_TO_VOXELSHAPE.get(state);
	}

	public boolean canPlaceAt(BlockState state, WorldView world, BlockPos pos) {
		BlockPos blockPos = pos.down();
		BlockState blockState = world.getBlockState(blockPos);
		return this.canRunOnTop(world, blockPos, blockState);
	}

	private boolean canRunOnTop(BlockView world, BlockPos pos, BlockState floor) {
		return floor.isSideSolidFullSquare(world, pos, Direction.UP);
	}

	@Override
	public BlockState rotate(BlockState state, BlockRotation rotation) {
		switch(rotation) {
		case CLOCKWISE_180:
			 return (BlockState)((BlockState)((BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_SOUTH))).with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_NORTH))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_EAST));
		case COUNTERCLOCKWISE_90:
			 return (BlockState)((BlockState)((BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_EAST))).with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_SOUTH))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_NORTH));
		case CLOCKWISE_90:
			 return (BlockState)((BlockState)((BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_NORTH))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_EAST))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_SOUTH));
		default:
			 return state;
		}
	}

 	@Override
	 public BlockState mirror(BlockState state, BlockMirror mirror) {
		switch(mirror) {
		case LEFT_RIGHT:
			 return (BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_SOUTH))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_NORTH));
		case FRONT_BACK:
			 return (BlockState)((BlockState)state.with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_EAST));
		default:
			 return super.mirror(state, mirror);
		}
	}

	private static boolean hasConnection(BlockState state) {
		return ((WDWireConnection)state.get(WIRE_CONNECTION_NORTH)).isConnected() || 
				((WDWireConnection)state.get(WIRE_CONNECTION_SOUTH)).isConnected() ||
				((WDWireConnection)state.get(WIRE_CONNECTION_EAST)).isConnected() ||
				((WDWireConnection)state.get(WIRE_CONNECTION_WEST)).isConnected();
	}

	private BlockState determineState(WorldAccess world, BlockState state, BlockPos pos){
		// Iterates every direction, and determines if there needs to be a connection set on that direction
		Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
		while(dirItr.hasNext()){
			Direction dir = dirItr.next();
			if (!state.get(DIRECTION_TO_WIRE_CONNECTION.get(dir)).isBroken())
				state = state.with(DIRECTION_TO_WIRE_CONNECTION.get(dir), determineConnection(world, pos, dir));
		}

		// This prevents 1 side connection from exsisting, as we don't want that since we don't want voxel-shapes like that. (We also don't support it)
		boolean NORTH = state.get(WIRE_CONNECTION_NORTH).isConnected();
		boolean SOUTH = state.get(WIRE_CONNECTION_SOUTH).isConnected();
		boolean EAST = state.get(WIRE_CONNECTION_EAST).isConnected();
		boolean WEST = state.get(WIRE_CONNECTION_WEST).isConnected();

		// This overwrites BROKE features, meaning this entire system is kinda useless as it will cause some weird looking situations. 
		if (NORTH && !EAST && !WEST && !state.get(WIRE_CONNECTION_SOUTH).isBroken())
			state = state.with(WIRE_CONNECTION_SOUTH, WDWireConnection.SIDE);
		else if (SOUTH && !EAST && !WEST && !state.get(WIRE_CONNECTION_NORTH).isBroken())
			state = state.with(WIRE_CONNECTION_NORTH, WDWireConnection.SIDE);
		else if (!NORTH && !SOUTH && EAST && !state.get(WIRE_CONNECTION_WEST).isBroken())
			state = state.with(WIRE_CONNECTION_WEST, WDWireConnection.SIDE);
		else if (!NORTH && !SOUTH && WEST && !state.get(WIRE_CONNECTION_EAST).isBroken())
			state = state.with(WIRE_CONNECTION_EAST, WDWireConnection.SIDE);

			//TODO: FIX n&s issue

		return state;
	}

	private WDWireConnection determineConnection(WorldAccess world, BlockPos pos, Direction direction){
		BlockPos offsetPos = pos.offset(direction);
		if((world.getBlockState(offsetPos.up()).isOf(this) && world.isAir(pos.up()))) // If offset+up is me && if up is air
			if(!world.getBlockState(offsetPos.up()).get(DIRECTION_TO_WIRE_CONNECTION.get(direction.getOpposite())).isBroken()) //if offset+up is not broken
				if(world.getBlockState(offsetPos).hasSidedTransparency())	
					return WDWireConnection.SIDE;
				else	
					return WDWireConnection.UP;
		if(world.getBlockState(offsetPos.down()).isOf(this) && world.isAir(offsetPos)) // if offset+down is me && if offset is air
			if(!world.getBlockState(offsetPos.down()).get(DIRECTION_TO_WIRE_CONNECTION.get(direction.getOpposite())).isBroken()) //if offset+down is not broken
				return WDWireConnection.SIDE;		
		if(world.getBlockState(offsetPos).isOf(this)) // if offset is me
			if(!world.getBlockState(offsetPos).get(DIRECTION_TO_WIRE_CONNECTION.get(direction.getOpposite())).isBroken()) // if offset is not broken
				return WDWireConnection.SIDE;
		return WDWireConnection.NONE;
	}

	@Override // Get the state I should place in, and update those around me (above/below) as needed
	public BlockState getPlacementState(ItemPlacementContext ctx){
		return determineState(ctx.getWorld(), this.getDefaultState(), ctx.getBlockPos());
	}

	@Override
	public void onPlaced(World world, BlockPos pos, BlockState state, @Nullable LivingEntity placer, ItemStack itemStack) {
		world.updateNeighbors(pos.offset(Direction.UP), state.getBlock());
		world.updateNeighbors(pos.offset(Direction.DOWN), state.getBlock());
	}

	@Override // This causes w/e issue that tracks deal with.
	public void onBroken(WorldAccess world, BlockPos pos, BlockState state) {
		world.updateNeighbors(pos.offset(Direction.UP), state.getBlock());
		world.updateNeighbors(pos.offset(Direction.DOWN), state.getBlock());
	}

	@Override // Called when I get told to update, so I should see what/why I need to update. I also use this to determine if I need to be drop or not.
	public void neighborUpdate(BlockState state, World world, BlockPos pos, Block block, BlockPos fromPos, boolean notify) {
		if (!world.isClient) {
			if(!state.canPlaceAt(world, pos)) {
				world.removeBlock(pos, false);
				this.onBroken(world, pos, state);
				dropStacks(state, world, pos);
			} else {
				world.setBlockState(pos, determineState(world, state, pos));
			}
		}
	}

	// Place block check disconnect (if block block connection)
	@Override // For each link, set to break and update the blocks around me (above and below)
	public ActionResult onUse(BlockState state, World world, BlockPos pos, PlayerEntity player, Hand hand, BlockHitResult hit){
		if(world.isClient)
			if(hasConnection(state) && !player.hasStackEquipped(EquipmentSlot.MAINHAND))
				return ActionResult.SUCCESS;
			else
				return ActionResult.PASS;
		if(player.abilities.allowModifyWorld)
			if(hasConnection(state) && !player.hasStackEquipped(EquipmentSlot.MAINHAND)) {
				Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
				while(dirItr.hasNext()){
					Direction dir = dirItr.next();
					if(state.get(DIRECTION_TO_WIRE_CONNECTION.get(dir)).isConnected())
						state = state.with(DIRECTION_TO_WIRE_CONNECTION.get(dir), WDWireConnection.BROKE);
				}
				world.setBlockState(pos, state);
				world.updateNeighbors(pos.down(), this); // Update those below me
				world.updateNeighbors(pos.up(), this); // Update those above me
				return ActionResult.SUCCESS;
			}
		return ActionResult.PASS;
	}

	static {
		WIRE_CONNECTION_NORTH = EnumProperty.of("north", WDWireConnection.class);
		WIRE_CONNECTION_EAST = EnumProperty.of("east", WDWireConnection.class);
		WIRE_CONNECTION_SOUTH = EnumProperty.of("south", WDWireConnection.class);
		WIRE_CONNECTION_WEST = EnumProperty.of("west", WDWireConnection.class);
		
		DIRECTION_TO_WIRE_CONNECTION = Maps.newEnumMap((Map)ImmutableMap.of(
			Direction.NORTH, WIRE_CONNECTION_NORTH, 
			Direction.EAST, WIRE_CONNECTION_EAST, 
			Direction.SOUTH, WIRE_CONNECTION_SOUTH, 
			Direction.WEST, WIRE_CONNECTION_WEST)
		);
		
		DOT_VOXELSHAPE = Block.createCuboidShape(3.0D, 0.0D, 3.0D, 13.0D, 1.0D, 13.0D);

		DIRECTION_VOXELSHAPE = Maps.newEnumMap((Map)ImmutableMap.of(
			Direction.NORTH, Block.createCuboidShape(3.0D, 0.0D, 0.0D, 13.0D, 1.0D, 13.0D), 
			Direction.SOUTH, Block.createCuboidShape(3.0D, 0.0D, 3.0D, 13.0D, 1.0D, 16.0D), 
			Direction.EAST, Block.createCuboidShape(3.0D, 0.0D, 3.0D, 16.0D, 1.0D, 13.0D), 
			Direction.WEST, Block.createCuboidShape(0.0D, 0.0D, 3.0D, 13.0D, 1.0D, 13.0D))
		);

		UP_VOXELSHAPE = Maps.newEnumMap((Map)ImmutableMap.of(
			Direction.NORTH, VoxelShapes.union((VoxelShape)DIRECTION_VOXELSHAPE.get(Direction.NORTH), Block.createCuboidShape(3.0D, 0.0D, 0.0D, 13.0D, 16.0D, 1.0D)), 
			Direction.SOUTH, VoxelShapes.union((VoxelShape)DIRECTION_VOXELSHAPE.get(Direction.SOUTH), Block.createCuboidShape(3.0D, 0.0D, 15.0D, 13.0D, 16.0D, 16.0D)), 
			Direction.EAST, VoxelShapes.union((VoxelShape)DIRECTION_VOXELSHAPE.get(Direction.EAST), Block.createCuboidShape(15.0D, 0.0D, 3.0D, 16.0D, 16.0D, 13.0D)), 
			Direction.WEST, VoxelShapes.union((VoxelShape)DIRECTION_VOXELSHAPE.get(Direction.WEST), Block.createCuboidShape(0.0D, 0.0D, 3.0D, 1.0D, 16.0D, 13.0D)))
		); 
	}
}